from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.validators import ActionValidators
from ai.backend.manager.services.agent.processors import AgentProcessors
from ai.backend.manager.services.agent.service import AgentService
from ai.backend.manager.services.app_config.processors import AppConfigProcessors
from ai.backend.manager.services.app_config.service import AppConfigService
from ai.backend.manager.services.artifact.processors import ArtifactProcessors
from ai.backend.manager.services.artifact.service import ArtifactService
from ai.backend.manager.services.artifact_registry.processors import ArtifactRegistryProcessors
from ai.backend.manager.services.artifact_registry.service import ArtifactRegistryService
from ai.backend.manager.services.artifact_revision.processors import ArtifactRevisionProcessors
from ai.backend.manager.services.artifact_revision.service import ArtifactRevisionService
from ai.backend.manager.services.audit_log.processors import AuditLogProcessors
from ai.backend.manager.services.audit_log.service import AuditLogService
from ai.backend.manager.services.auth.processors import AuthProcessors
from ai.backend.manager.services.auth.service import AuthService
from ai.backend.manager.services.container_registry.processors import ContainerRegistryProcessors
from ai.backend.manager.services.container_registry.service import ContainerRegistryService
from ai.backend.manager.services.deployment.processors import DeploymentProcessors
from ai.backend.manager.services.deployment.service import DeploymentService
from ai.backend.manager.services.domain.processors import DomainProcessors
from ai.backend.manager.services.domain.service import DomainService
from ai.backend.manager.services.dotfile.processors import DotfileProcessors
from ai.backend.manager.services.dotfile.service import DotfileService
from ai.backend.manager.services.error_log.processors import ErrorLogProcessors
from ai.backend.manager.services.error_log.service import ErrorLogService
from ai.backend.manager.services.etcd_config.processors import EtcdConfigProcessors
from ai.backend.manager.services.etcd_config.service import EtcdConfigService
from ai.backend.manager.services.events.processors import EventsProcessors
from ai.backend.manager.services.events.service import EventsService
from ai.backend.manager.services.export.processors import ExportProcessors
from ai.backend.manager.services.export.service import ExportService
from ai.backend.manager.services.fair_share.processors import FairShareProcessors
from ai.backend.manager.services.fair_share.service import FairShareService
from ai.backend.manager.services.group.processors import GroupProcessors
from ai.backend.manager.services.group.service import GroupService
from ai.backend.manager.services.image.processors import ImageProcessors
from ai.backend.manager.services.image.service import ImageService
from ai.backend.manager.services.keypair_resource_policy.processors import (
    KeypairResourcePolicyProcessors,
)
from ai.backend.manager.services.keypair_resource_policy.service import KeypairResourcePolicyService
from ai.backend.manager.services.manager_admin.processors import ManagerAdminProcessors
from ai.backend.manager.services.manager_admin.service import ManagerAdminService
from ai.backend.manager.services.metric.processors.utilization_metric import (
    UtilizationMetricProcessors,
)
from ai.backend.manager.services.metric.root_service import UtilizationMetricService
from ai.backend.manager.services.model_serving.processors.auto_scaling import (
    ModelServingAutoScalingProcessors,
)
from ai.backend.manager.services.model_serving.processors.model_serving import (
    ModelServingProcessors,
)
from ai.backend.manager.services.model_serving.services.auto_scaling import AutoScalingService
from ai.backend.manager.services.model_serving.services.model_serving import (
    ModelServingService,
)
from ai.backend.manager.services.notification.processors import NotificationProcessors
from ai.backend.manager.services.notification.service import NotificationService
from ai.backend.manager.services.object_storage.processors import ObjectStorageProcessors
from ai.backend.manager.services.object_storage.service import ObjectStorageService
from ai.backend.manager.services.permission_contoller.processors import (
    PermissionControllerProcessors,
)
from ai.backend.manager.services.permission_contoller.service import PermissionControllerService
from ai.backend.manager.services.processors import ProcessorArgs, Processors, ServiceArgs, Services
from ai.backend.manager.services.project_resource_policy.processors import (
    ProjectResourcePolicyProcessors,
)
from ai.backend.manager.services.project_resource_policy.service import ProjectResourcePolicyService
from ai.backend.manager.services.prometheus_query_preset.processors import (
    PrometheusQueryPresetProcessors,
)
from ai.backend.manager.services.prometheus_query_preset.service import (
    PrometheusQueryPresetService,
)
from ai.backend.manager.services.resource_preset.processors import ResourcePresetProcessors
from ai.backend.manager.services.resource_preset.service import ResourcePresetService
from ai.backend.manager.services.resource_slot.processors import ResourceSlotProcessors
from ai.backend.manager.services.resource_slot.service import ResourceSlotService
from ai.backend.manager.services.resource_usage.processors import ResourceUsageProcessors
from ai.backend.manager.services.resource_usage.service import ResourceUsageService
from ai.backend.manager.services.scaling_group.processors import ScalingGroupProcessors
from ai.backend.manager.services.scaling_group.service import ScalingGroupService
from ai.backend.manager.services.scheduling_history.processors import SchedulingHistoryProcessors
from ai.backend.manager.services.scheduling_history.service import SchedulingHistoryService
from ai.backend.manager.services.service_catalog.processors import ServiceCatalogProcessors
from ai.backend.manager.services.service_catalog.service import ServiceCatalogService
from ai.backend.manager.services.session.processors import SessionProcessors
from ai.backend.manager.services.session.service import SessionService, SessionServiceArgs
from ai.backend.manager.services.storage_namespace.processors import StorageNamespaceProcessors
from ai.backend.manager.services.storage_namespace.service import StorageNamespaceService
from ai.backend.manager.services.stream.processors import StreamProcessors
from ai.backend.manager.services.stream.service import StreamService
from ai.backend.manager.services.template.processors import TemplateProcessors
from ai.backend.manager.services.template.service import TemplateService
from ai.backend.manager.services.user.processors import UserProcessors
from ai.backend.manager.services.user.service import UserService
from ai.backend.manager.services.user_resource_policy.processors import UserResourcePolicyProcessors
from ai.backend.manager.services.user_resource_policy.service import UserResourcePolicyService
from ai.backend.manager.services.vfolder.processors import (
    VFolderFileProcessors,
    VFolderInviteProcessors,
    VFolderProcessors,
    VFolderSharingProcessors,
)
from ai.backend.manager.services.vfolder.services.file import VFolderFileService
from ai.backend.manager.services.vfolder.services.invite import VFolderInviteService
from ai.backend.manager.services.vfolder.services.sharing import VFolderSharingService
from ai.backend.manager.services.vfolder.services.vfolder import VFolderService
from ai.backend.manager.services.vfs_storage.processors import VFSStorageProcessors
from ai.backend.manager.services.vfs_storage.service import VFSStorageService


def create_services(args: ServiceArgs) -> Services:
    repositories = args.repositories
    return Services(
        agent=AgentService(
            args.etcd,
            args.agent_registry,
            args.config_provider,
            repositories.agent.repository,
            repositories.scheduler.repository,
            args.hook_plugin_ctx,
            args.event_producer,
            args.agent_cache,
        ),
        app_config=AppConfigService(
            app_config_repository=repositories.app_config.repository,
        ),
        domain=DomainService(repositories.domain.repository),
        dotfile=DotfileService(
            repository=repositories.dotfile.repository,
        ),
        error_log=ErrorLogService(
            repository=repositories.error_log.repository,
        ),
        etcd_config=EtcdConfigService(
            repository=repositories.etcd_config.repository,
            config_provider=args.config_provider,
            etcd=args.etcd,
            valkey_stat=args.valkey_stat_client,
        ),
        export=ExportService(
            repository=repositories.export.repository,
        ),
        fair_share=FairShareService(
            repository=repositories.fair_share.repository,
        ),
        group=GroupService(
            args.storage_manager,
            args.config_provider,
            args.valkey_stat_client,
            repositories.group,
        ),
        user=UserService(
            args.storage_manager,
            args.valkey_stat_client,
            args.agent_registry,
            repositories.user.repository,
        ),
        image=ImageService(
            args.agent_registry, repositories.image.repository, args.config_provider
        ),
        container_registry=ContainerRegistryService(
            args.db,
            repositories.container_registry.repository,
            quota_service=args.registry_quota_service,
        ),
        vfolder=VFolderService(
            args.config_provider,
            args.etcd,
            args.storage_manager,
            args.background_task_manager,
            repositories.vfolder.repository,
            repositories.user.repository,
            args.valkey_stat_client,
        ),
        vfolder_file=VFolderFileService(
            args.config_provider,
            args.storage_manager,
            repositories.vfolder.repository,
            repositories.user.repository,
        ),
        vfolder_invite=VFolderInviteService(
            args.config_provider,
            repositories.vfolder.repository,
            repositories.user.repository,
        ),
        vfolder_sharing=VFolderSharingService(
            args.config_provider,
            repositories.vfolder.repository,
            repositories.user.repository,
        ),
        session=SessionService(
            SessionServiceArgs(
                agent_registry=args.agent_registry,
                event_fetcher=args.event_fetcher,
                background_task_manager=args.background_task_manager,
                event_hub=args.event_hub,
                error_monitor=args.error_monitor,
                idle_checker_host=args.idle_checker_host,
                session_repository=repositories.session.repository,
                scheduling_controller=args.scheduling_controller,
                appproxy_client_pool=args.appproxy_client_pool,
            )
        ),
        keypair_resource_policy=KeypairResourcePolicyService(
            repositories.keypair_resource_policy.repository
        ),
        manager_admin=ManagerAdminService(
            repository=repositories.manager_admin.repository,
            config_provider=args.config_provider,
            etcd=args.etcd,
            db=args.db,
            valkey_stat=args.valkey_stat_client,
        ),
        user_resource_policy=UserResourcePolicyService(
            repositories.user_resource_policy.repository
        ),
        project_resource_policy=ProjectResourcePolicyService(
            repositories.project_resource_policy.repository
        ),
        prometheus_query_preset=PrometheusQueryPresetService(
            repository=repositories.prometheus_query_preset.repository,
            prometheus_client=args.prometheus_client,
            default_timewindow=args.config_provider.config.metric.timewindow,
        ),
        resource_preset=ResourcePresetService(
            repositories.resource_preset.repository,
        ),
        resource_slot=ResourceSlotService(repositories.resource_slot.repository),
        resource_usage=ResourceUsageService(
            repository=repositories.resource_usage_history.repository,
        ),
        scaling_group=ScalingGroupService(
            repositories.scaling_group.repository,
            appproxy_client_pool=args.appproxy_client_pool,
        ),
        utilization_metric=UtilizationMetricService(
            args.prometheus_client,
            args.config_provider.config.metric.timewindow,
            repositories.metric.repository,
        ),
        model_serving=ModelServingService(
            agent_registry=args.agent_registry,
            background_task_manager=args.background_task_manager,
            event_dispatcher=args.event_dispatcher,
            event_hub=args.event_hub,
            storage_manager=args.storage_manager,
            config_provider=args.config_provider,
            valkey_live=args.valkey_live,
            repository=repositories.model_serving.repository,
            deployment_repository=args.deployment_controller._deployment_repository,
            deployment_controller=args.deployment_controller,
            scheduling_controller=args.scheduling_controller,
            revision_generator_registry=args.revision_generator_registry,
        ),
        model_serving_auto_scaling=AutoScalingService(
            repository=repositories.model_serving.repository,
        ),
        auth=AuthService(
            hook_plugin_ctx=args.hook_plugin_ctx,
            auth_repository=repositories.auth.repository,
            config_provider=args.config_provider,
        ),
        notification=NotificationService(
            repository=repositories.notification.repository,
            notification_center=args.notification_center,
        ),
        object_storage=ObjectStorageService(
            artifact_repository=repositories.artifact.repository,
            object_storage_repository=repositories.object_storage.repository,
            storage_namespace_repository=repositories.storage_namespace.repository,
            storage_manager=args.storage_manager,
            config_provider=args.config_provider,
        ),
        permission_controller=PermissionControllerService(
            repository=repositories.permission_controller.repository,
        ),
        vfs_storage=VFSStorageService(
            vfs_storage_repository=repositories.vfs_storage.repository,
            storage_manager=args.storage_manager,
        ),
        artifact=ArtifactService(
            artifact_repository=repositories.artifact.repository,
            artifact_registry_repository=repositories.artifact_registry.repository,
            storage_manager=args.storage_manager,
            object_storage_repository=repositories.object_storage.repository,
            vfs_storage_repository=repositories.vfs_storage.repository,
            huggingface_registry_repository=repositories.huggingface_registry.repository,
            config_provider=args.config_provider,
            reservoir_registry_repository=repositories.reservoir_registry.repository,
        ),
        artifact_revision=ArtifactRevisionService(
            artifact_repository=repositories.artifact.repository,
            artifact_registry_repository=repositories.artifact_registry.repository,
            storage_manager=args.storage_manager,
            object_storage_repository=repositories.object_storage.repository,
            vfs_storage_repository=repositories.vfs_storage.repository,
            storage_namespace_repository=repositories.storage_namespace.repository,
            huggingface_registry_repository=repositories.huggingface_registry.repository,
            reservoir_registry_repository=repositories.reservoir_registry.repository,
            vfolder_repository=repositories.vfolder.repository,
            config_provider=args.config_provider,
            valkey_artifact_client=args.valkey_artifact_client,
            background_task_manager=args.background_task_manager,
        ),
        artifact_registry=ArtifactRegistryService(
            repositories.huggingface_registry.repository,
            repositories.reservoir_registry.repository,
            repositories.artifact_registry.repository,
        ),
        deployment=DeploymentService(
            args.deployment_controller,
            args.deployment_controller._deployment_repository,
            args.revision_generator_registry,
        ),
        storage_namespace=StorageNamespaceService(repositories.storage_namespace.repository),
        audit_log=AuditLogService(repositories.audit_log.repository),
        scheduling_history=SchedulingHistoryService(repositories.scheduling_history.repository),
        service_catalog=ServiceCatalogService(args.db),
        template=TemplateService(
            repository=repositories.template.repository,
        ),
        stream=StreamService(
            repository=repositories.stream.repository,
            registry=args.agent_registry,
            valkey_live=args.valkey_live,
            idle_checker_host=args.idle_checker_host,
            etcd=args.etcd,
        ),
        events=EventsService(
            repository=repositories.events.repository,
            db=args.db,
        ),
    )


def create_processors(
    args: ProcessorArgs,
    action_monitors: list[ActionMonitor],
    validators: ActionValidators,
) -> Processors:
    services = create_services(args.service_args)
    return Processors(
        agent=AgentProcessors(services.agent, action_monitors, validators),
        app_config=AppConfigProcessors(services.app_config, action_monitors, validators),
        domain=DomainProcessors(services.domain, action_monitors, validators),
        dotfile=DotfileProcessors(services.dotfile, action_monitors, validators),
        error_log=ErrorLogProcessors(services.error_log, action_monitors, validators),
        etcd_config=EtcdConfigProcessors(services.etcd_config, action_monitors, validators),
        export=ExportProcessors(services.export, action_monitors, validators),
        fair_share=FairShareProcessors(services.fair_share, action_monitors, validators),
        group=GroupProcessors(services.group, action_monitors, validators),
        user=UserProcessors(services.user, action_monitors, validators),
        image=ImageProcessors(services.image, action_monitors, validators),
        container_registry=ContainerRegistryProcessors(
            services.container_registry, action_monitors, validators
        ),
        vfolder=VFolderProcessors(services.vfolder, action_monitors, validators),
        vfolder_file=VFolderFileProcessors(services.vfolder_file, action_monitors, validators),
        vfolder_invite=VFolderInviteProcessors(
            services.vfolder_invite, action_monitors, validators
        ),
        vfolder_sharing=VFolderSharingProcessors(
            services.vfolder_sharing, action_monitors, validators
        ),
        session=SessionProcessors(services.session, action_monitors, validators),
        keypair_resource_policy=KeypairResourcePolicyProcessors(
            services.keypair_resource_policy, action_monitors, validators
        ),
        manager_admin=ManagerAdminProcessors(services.manager_admin, action_monitors, validators),
        user_resource_policy=UserResourcePolicyProcessors(
            services.user_resource_policy, action_monitors, validators
        ),
        project_resource_policy=ProjectResourcePolicyProcessors(
            services.project_resource_policy, action_monitors, validators
        ),
        prometheus_query_preset=PrometheusQueryPresetProcessors(
            services.prometheus_query_preset, action_monitors, validators
        ),
        resource_preset=ResourcePresetProcessors(
            services.resource_preset, action_monitors, validators
        ),
        resource_slot=ResourceSlotProcessors(services.resource_slot, action_monitors, validators),
        resource_usage=ResourceUsageProcessors(
            services.resource_usage, action_monitors, validators
        ),
        scaling_group=ScalingGroupProcessors(services.scaling_group, action_monitors, validators),
        utilization_metric=UtilizationMetricProcessors(
            services.utilization_metric, action_monitors, validators
        ),
        model_serving=ModelServingProcessors(services.model_serving, action_monitors, validators),
        model_serving_auto_scaling=ModelServingAutoScalingProcessors(
            services.model_serving_auto_scaling, action_monitors, validators
        ),
        auth=AuthProcessors(services.auth, action_monitors, validators),
        notification=NotificationProcessors(services.notification, action_monitors, validators),
        object_storage=ObjectStorageProcessors(
            services.object_storage, action_monitors, validators
        ),
        permission_controller=PermissionControllerProcessors(
            services.permission_controller, action_monitors, validators
        ),
        vfs_storage=VFSStorageProcessors(services.vfs_storage, action_monitors, validators),
        artifact=ArtifactProcessors(services.artifact, action_monitors, validators),
        artifact_registry=ArtifactRegistryProcessors(
            services.artifact_registry, action_monitors, validators
        ),
        artifact_revision=ArtifactRevisionProcessors(
            services.artifact_revision, action_monitors, validators
        ),
        deployment=DeploymentProcessors(services.deployment, action_monitors, validators),
        storage_namespace=StorageNamespaceProcessors(
            services.storage_namespace, action_monitors, validators
        ),
        audit_log=AuditLogProcessors(services.audit_log, [], validators),
        scheduling_history=SchedulingHistoryProcessors(
            services.scheduling_history, action_monitors, validators
        ),
        service_catalog=ServiceCatalogProcessors(
            services.service_catalog, action_monitors, validators
        ),
        template=TemplateProcessors(services.template, action_monitors, validators),
        stream=StreamProcessors(services.stream, action_monitors),
        events=EventsProcessors(
            services.events,
            action_monitors,
            event_hub=args.event_hub,
            event_fetcher=args.event_fetcher,
        ),
    )



from __future__ import annotations

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any, override
from uuid import UUID

from ai.backend.common.exception import ScalingGroupConflict
from ai.backend.common.types import AccessKey
from ai.backend.manager.errors.repository import UniqueConstraintViolationError
from ai.backend.manager.models.scaling_group import (
    ScalingGroupForDomainRow,
    ScalingGroupForKeypairsRow,
    ScalingGroupForProjectRow,
    ScalingGroupOpts,
    ScalingGroupRow,
)
from ai.backend.manager.models.scaling_group.types import FairShareScalingGroupSpec
from ai.backend.manager.repositories.base.creator import CreatorSpec
from ai.backend.manager.repositories.base.types import IntegrityErrorCheck


@dataclass
class ScalingGroupCreatorSpec(CreatorSpec[ScalingGroupRow]):
    """CreatorSpec for scaling group."""

    name: str
    driver: str
    scheduler: str
    description: str | None = None
    is_active: bool = True
    is_public: bool = True
    wsproxy_addr: str | None = None
    wsproxy_api_token: str | None = None
    driver_opts: Mapping[str, Any] = field(default_factory=dict)
    scheduler_opts: ScalingGroupOpts | None = None
    use_host_network: bool = False
    fair_share_spec: FairShareScalingGroupSpec | None = None

    @property
    @override
    def integrity_error_checks(self) -> Sequence[IntegrityErrorCheck]:
        return (
            IntegrityErrorCheck(
                violation_type=UniqueConstraintViolationError,
                error=ScalingGroupConflict(f"Duplicate scaling group name: {self.name}"),
            ),
        )

    @override
    def build_row(self) -> ScalingGroupRow:
        return ScalingGroupRow(
            name=self.name,
            description=self.description,
            is_active=self.is_active,
            is_public=self.is_public,
            wsproxy_addr=self.wsproxy_addr,
            wsproxy_api_token=self.wsproxy_api_token,
            driver=self.driver,
            driver_opts=dict(self.driver_opts),
            scheduler=self.scheduler,
            scheduler_opts=self.scheduler_opts if self.scheduler_opts else ScalingGroupOpts(),
            use_host_network=self.use_host_network,
            fair_share_spec=self.fair_share_spec,
        )


@dataclass
class ScalingGroupForDomainCreatorSpec(CreatorSpec[ScalingGroupForDomainRow]):
    """CreatorSpec for associating a scaling group with a domain."""

    scaling_group: str
    domain: str

    @override
    def build_row(self) -> ScalingGroupForDomainRow:
        return ScalingGroupForDomainRow(
            scaling_group=self.scaling_group,
            domain=self.domain,
        )


@dataclass
class ScalingGroupForKeypairsCreatorSpec(CreatorSpec[ScalingGroupForKeypairsRow]):
    """CreatorSpec for associating a scaling group with a keypair."""

    scaling_group: str
    access_key: AccessKey

    @override
    def build_row(self) -> ScalingGroupForKeypairsRow:
        return ScalingGroupForKeypairsRow(
            scaling_group=self.scaling_group,
            access_key=self.access_key,
        )


@dataclass
class ScalingGroupForProjectCreatorSpec(CreatorSpec[ScalingGroupForProjectRow]):
    """CreatorSpec for associating a scaling group with a project (user group)."""

    scaling_group: str
    project: UUID

    @override
    def build_row(self) -> ScalingGroupForProjectRow:
        return ScalingGroupForProjectRow(
            scaling_group=self.scaling_group,
            group=self.project,
        )



"""Revision fetcher functions."""

from __future__ import annotations

from functools import lru_cache
from uuid import UUID

from strawberry import Info
from strawberry.relay import PageInfo

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.deployment.types.revision import (
    ModelRevision,
    ModelRevisionConnection,
    ModelRevisionEdge,
    ModelRevisionFilter,
    ModelRevisionOrderBy,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.deployment_revision.row import DeploymentRevisionRow
from ai.backend.manager.repositories.base import QueryCondition
from ai.backend.manager.repositories.deployment.options import (
    RevisionConditions,
    RevisionOrders,
)
from ai.backend.manager.services.deployment.actions.model_revision.search_revisions import (
    SearchRevisionsAction,
)


@lru_cache(maxsize=1)
def get_revision_pagination_spec() -> PaginationSpec:
    return PaginationSpec(
        forward_order=RevisionOrders.created_at(ascending=False),
        backward_order=RevisionOrders.created_at(ascending=True),
        forward_condition_factory=RevisionConditions.by_cursor_forward,
        backward_condition_factory=RevisionConditions.by_cursor_backward,
        tiebreaker_order=DeploymentRevisionRow.id.asc(),
    )


async def fetch_revisions(
    info: Info[StrawberryGQLContext],
    filter: ModelRevisionFilter | None = None,
    order_by: list[ModelRevisionOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
    base_conditions: list[QueryCondition] | None = None,
) -> ModelRevisionConnection:
    """Fetch revisions with optional filtering, ordering, and pagination.

    Args:
        info: GraphQL context info
        filter: Optional filter criteria
        order_by: Optional ordering specification
        before/after/first/last: Cursor-based pagination parameters
        limit/offset: Offset-based pagination parameters
        base_conditions: Additional conditions to prepend (e.g., deployment_id filter)
    """
    processor = info.context.processors.deployment

    # Build querier using gql_adapter
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        get_revision_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=base_conditions,
    )

    action_result = await processor.search_revisions.wait_for_complete(
        SearchRevisionsAction(querier=querier)
    )

    nodes = [ModelRevision.from_dataclass(data) for data in action_result.data]
    edges = [ModelRevisionEdge(node=node, cursor=encode_cursor(str(node.id))) for node in nodes]

    return ModelRevisionConnection(
        edges=edges,
        page_info=PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )


async def fetch_revision(
    info: Info[StrawberryGQLContext],
    revision_id: UUID,
) -> ModelRevision | None:
    """Fetch a specific revision by ID."""
    revision_data = await info.context.data_loaders.revision_loader.load(revision_id)
    if revision_data is None:
        return None
    return ModelRevision.from_dataclass(revision_data)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.manager.agent_cache import AgentRPCCache
from ai.backend.manager.clients.agent import AgentClientPool, AgentPoolSpec


@dataclass
class AgentClientPoolInput:
    """Input required for agent client pool setup."""

    agent_cache: AgentRPCCache


class AgentClientPoolDependency(
    NonMonitorableDependencyProvider[AgentClientPoolInput, AgentClientPool],
):
    """Provides AgentClientPool lifecycle management."""

    @property
    def stage_name(self) -> str:
        return "agent-client-pool"

    @asynccontextmanager
    async def provide(self, setup_input: AgentClientPoolInput) -> AsyncIterator[AgentClientPool]:
        """Initialize and provide an agent client pool.

        Args:
            setup_input: Input containing agent RPC cache

        Yields:
            Initialized AgentClientPool
        """
        pool = AgentClientPool(
            setup_input.agent_cache,
            AgentPoolSpec(
                health_check_interval=30.0,
                failure_threshold=3,
                recovery_timeout=60.0,
            ),
        )
        try:
            yield pool
        finally:
            await pool.close()



"""
Adapter to convert ResourceSlotType DTOs to repository Querier objects.
"""

from __future__ import annotations

from ai.backend.common.dto.manager.resource_slot.request import (
    OrderDirection,
    ResourceSlotTypeFilter,
    ResourceSlotTypeOrder,
    ResourceSlotTypeOrderField,
    SearchResourceSlotTypesRequest,
)
from ai.backend.common.dto.manager.resource_slot.response import (
    NumberFormatDTO,
    ResourceSlotTypeDTO,
)
from ai.backend.manager.api.rest.adapter import BaseFilterAdapter
from ai.backend.manager.data.resource_slot.types import ResourceSlotTypeData
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    OffsetPagination,
    QueryCondition,
    QueryOrder,
)
from ai.backend.manager.repositories.resource_slot.query import QueryConditions, QueryOrders

__all__ = ("ResourceSlotAdapter",)


class ResourceSlotAdapter(BaseFilterAdapter):
    """Adapter for converting resource slot type requests to repository queries."""

    def convert_to_dto(self, data: ResourceSlotTypeData) -> ResourceSlotTypeDTO:
        """Convert ResourceSlotTypeData to DTO."""
        return ResourceSlotTypeDTO(
            slot_name=data.slot_name,
            slot_type=data.slot_type,
            display_name=data.display_name,
            description=data.description,
            display_unit=data.display_unit,
            display_icon=data.display_icon,
            number_format=NumberFormatDTO(
                binary=data.number_format.binary,
                round_length=data.number_format.round_length,
            ),
            rank=data.rank,
        )

    def build_querier(self, request: SearchResourceSlotTypesRequest) -> BatchQuerier:
        """Build a BatchQuerier from search request."""
        conditions = self._convert_filter(request.filter) if request.filter else []
        orders = [self._convert_order(o) for o in request.order] if request.order else []

        return BatchQuerier(
            conditions=conditions,
            orders=orders,
            pagination=OffsetPagination(limit=request.limit, offset=request.offset),
        )

    def _convert_filter(self, filter: ResourceSlotTypeFilter) -> list[QueryCondition]:
        """Convert resource slot type filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        if filter.slot_name is not None:
            condition = self.convert_string_filter(
                filter.slot_name,
                contains_factory=QueryConditions.by_slot_name_contains,
                equals_factory=QueryConditions.by_slot_name_equals,
                starts_with_factory=QueryConditions.by_slot_name_starts_with,
                ends_with_factory=QueryConditions.by_slot_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.slot_type is not None:
            condition = self.convert_string_filter(
                filter.slot_type,
                contains_factory=QueryConditions.by_slot_type_contains,
                equals_factory=QueryConditions.by_slot_type_equals,
                starts_with_factory=QueryConditions.by_slot_type_starts_with,
                ends_with_factory=QueryConditions.by_slot_type_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.display_name is not None:
            condition = self.convert_string_filter(
                filter.display_name,
                contains_factory=QueryConditions.by_display_name_contains,
                equals_factory=QueryConditions.by_display_name_equals,
                starts_with_factory=QueryConditions.by_display_name_starts_with,
                ends_with_factory=QueryConditions.by_display_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        return conditions

    def _convert_order(self, order: ResourceSlotTypeOrder) -> QueryOrder:
        """Convert resource slot type order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        if order.field == ResourceSlotTypeOrderField.SLOT_NAME:
            return QueryOrders.slot_name(ascending=ascending)
        if order.field == ResourceSlotTypeOrderField.RANK:
            return QueryOrders.rank(ascending=ascending)
        if order.field == ResourceSlotTypeOrderField.DISPLAY_NAME:
            return QueryOrders.display_name(ascending=ascending)
        raise ValueError(f"Unknown order field: {order.field}")



"""REST API handler for Prometheus Query Preset operations."""

from __future__ import annotations

from http import HTTPStatus

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam
from ai.backend.common.dto.manager.prometheus_query_preset import (
    CreateQueryDefinitionRequest,
    CreateQueryDefinitionResponse,
    DeleteQueryDefinitionResponse,
    ExecuteQueryDefinitionRequest,
    ExecuteQueryDefinitionResponse,
    GetQueryDefinitionResponse,
    ModifyQueryDefinitionRequest,
    ModifyQueryDefinitionResponse,
    PaginationInfo,
    QueryDefinitionExecuteData,
    QueryDefinitionIdPathParam,
    SearchQueryDefinitionsRequest,
    SearchQueryDefinitionsResponse,
)
from ai.backend.manager.data.prometheus_query_preset import (
    ExecutePresetOptions,
)
from ai.backend.manager.models.prometheus_query_preset import PrometheusQueryPresetRow
from ai.backend.manager.repositories.base import (
    Creator,
)
from ai.backend.manager.repositories.prometheus_query_preset.creators import (
    PrometheusQueryPresetCreatorSpec,
)
from ai.backend.manager.services.prometheus_query_preset.actions.create import (
    CreatePresetAction,
)
from ai.backend.manager.services.prometheus_query_preset.actions.delete import (
    DeletePresetAction,
)
from ai.backend.manager.services.prometheus_query_preset.actions.execute_preset import (
    ExecutePresetAction,
)
from ai.backend.manager.services.prometheus_query_preset.actions.get import (
    GetPresetAction,
)
from ai.backend.manager.services.prometheus_query_preset.actions.modify import (
    ModifyPresetAction,
)
from ai.backend.manager.services.prometheus_query_preset.actions.search import (
    SearchPresetsAction,
)
from ai.backend.manager.services.prometheus_query_preset.processors import (
    PrometheusQueryPresetProcessors,
)

from .adapter import PrometheusQueryPresetAdapter


class PrometheusQueryPresetHandler:
    """REST API handler for Prometheus Query Preset CRUD and execution."""

    def __init__(
        self,
        *,
        processor: PrometheusQueryPresetProcessors,
    ) -> None:
        self._processor = processor
        self._adapter = PrometheusQueryPresetAdapter()

    async def create_preset(
        self,
        body: BodyParam[CreateQueryDefinitionRequest],
    ) -> APIResponse:
        """Create a new prometheus query preset."""
        creator: Creator[PrometheusQueryPresetRow] = Creator(
            spec=PrometheusQueryPresetCreatorSpec(
                name=body.parsed.name,
                metric_name=body.parsed.metric_name,
                query_template=body.parsed.query_template,
                time_window=body.parsed.time_window,
                filter_labels=body.parsed.options.filter_labels,
                group_labels=body.parsed.options.group_labels,
            )
        )
        action_result = await self._processor.create_preset.wait_for_complete(
            CreatePresetAction(creator=creator)
        )
        resp = CreateQueryDefinitionResponse(
            item=self._adapter.convert_to_dto(action_result.preset)
        )
        return APIResponse.build(status_code=HTTPStatus.CREATED, response_model=resp)

    async def search_presets(
        self,
        body: BodyParam[SearchQueryDefinitionsRequest],
    ) -> APIResponse:
        """Search presets with filters, orders, and pagination."""
        querier = self._adapter.build_querier(body.parsed)
        action_result = await self._processor.search_presets.wait_for_complete(
            SearchPresetsAction(querier=querier)
        )
        resp = SearchQueryDefinitionsResponse(
            items=[
                self._adapter.convert_to_dto(preset_data) for preset_data in action_result.items
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def get_preset(
        self,
        path: PathParam[QueryDefinitionIdPathParam],
    ) -> APIResponse:
        """Get a preset by ID."""
        action_result = await self._processor.get_preset.wait_for_complete(
            GetPresetAction(preset_id=path.parsed.id)
        )
        resp = GetQueryDefinitionResponse(item=self._adapter.convert_to_dto(action_result.preset))
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def modify_preset(
        self,
        path: PathParam[QueryDefinitionIdPathParam],
        body: BodyParam[ModifyQueryDefinitionRequest],
    ) -> APIResponse:
        """Modify a preset."""
        updater = self._adapter.build_updater(body.parsed, path.parsed.id)
        action_result = await self._processor.modify_preset.wait_for_complete(
            ModifyPresetAction(preset_id=path.parsed.id, updater=updater)
        )
        resp = ModifyQueryDefinitionResponse(
            item=self._adapter.convert_to_dto(action_result.preset)
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def delete_preset(
        self,
        path: PathParam[QueryDefinitionIdPathParam],
    ) -> APIResponse:
        """Delete a preset."""
        action_result = await self._processor.delete_preset.wait_for_complete(
            DeletePresetAction(preset_id=path.parsed.id)
        )
        resp = DeleteQueryDefinitionResponse(id=action_result.preset_id)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def execute_preset(
        self,
        path: PathParam[QueryDefinitionIdPathParam],
        body: BodyParam[ExecuteQueryDefinitionRequest],
    ) -> APIResponse:
        """Execute a preset with given parameters."""
        filter_labels = {entry.key: entry.value for entry in body.parsed.options.filter_labels}
        options = ExecutePresetOptions(
            filter_labels=filter_labels,
            group_labels=body.parsed.options.group_labels,
        )

        action_result = await self._processor.execute_preset.wait_for_complete(
            ExecutePresetAction(
                preset_id=path.parsed.id,
                options=options,
                time_window=body.parsed.time_window,
                time_range=body.parsed.time_range,
            )
        )

        prom_response = action_result.response
        result_items = [
            self._adapter.convert_metric_response(metric_response)
            for metric_response in prom_response.data.result
        ]
        resp = ExecuteQueryDefinitionResponse(
            status=prom_response.status,
            data=QueryDefinitionExecuteData(
                result_type=prom_response.data.result_type,
                result=result_items,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)



from collections.abc import Mapping
from typing import Any, override

from aiohttp_sse import EventSourceResponse

from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.events.hub.propagators.session import SessionEventPropagator
from ai.backend.manager.services.events.actions.resolve_group_for_events import (
    ResolveGroupForEventsAction,
    ResolveGroupForEventsActionResult,
)
from ai.backend.manager.services.events.actions.resolve_session_for_events import (
    ResolveSessionForEventsAction,
    ResolveSessionForEventsActionResult,
)
from ai.backend.manager.services.events.service import EventsService


class EventsProcessors(AbstractProcessorPackage):
    _service: EventsService
    event_hub: EventHub
    event_fetcher: EventFetcher

    resolve_session: ActionProcessor[
        ResolveSessionForEventsAction, ResolveSessionForEventsActionResult
    ]
    resolve_group: ActionProcessor[ResolveGroupForEventsAction, ResolveGroupForEventsActionResult]

    def __init__(
        self,
        service: EventsService,
        action_monitors: list[ActionMonitor],
        *,
        event_hub: EventHub,
        event_fetcher: EventFetcher,
    ) -> None:
        self._service = service
        self.event_hub = event_hub
        self.event_fetcher = event_fetcher
        self.resolve_session = ActionProcessor(service.resolve_session_for_events, action_monitors)
        self.resolve_group = ActionProcessor(service.resolve_group_for_events, action_monitors)

    def create_session_propagator(
        self,
        response: EventSourceResponse,
        filters: Mapping[str, Any],
    ) -> SessionEventPropagator:
        return self._service.create_session_propagator(response, filters)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            ResolveSessionForEventsAction.spec(),
            ResolveGroupForEventsAction.spec(),
        ]



"""Prometheus query preset GQL resolvers."""

from .mutation import (
    admin_create_prometheus_query_preset,
    admin_delete_prometheus_query_preset,
    admin_modify_prometheus_query_preset,
)
from .query import (
    admin_prometheus_query_preset,
    admin_prometheus_query_preset_result,
    admin_prometheus_query_presets,
)

__all__ = [
    # Queries
    "admin_prometheus_query_preset",
    "admin_prometheus_query_presets",
    "admin_prometheus_query_preset_result",
    # Mutations
    "admin_create_prometheus_query_preset",
    "admin_modify_prometheus_query_preset",
    "admin_delete_prometheus_query_preset",
]



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import TemplateAction


@dataclass
class ListTaskTemplatesAction(TemplateAction):
    """Action to list all active task templates."""

    user_uuid: uuid.UUID

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class ListTaskTemplatesActionResult(BaseActionResult):
    """Result of listing task templates."""

    entries: list[dict[str, Any]] = field(default_factory=list)

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import ScalingGroupAction


@dataclass
class GetWsproxyVersionAction(ScalingGroupAction):
    """Action to get wsproxy version for a scaling group."""

    scaling_group_name: str
    domain_name: str
    group: str
    access_key: str

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET

    @override
    def entity_id(self) -> str | None:
        return self.scaling_group_name


@dataclass
class GetWsproxyVersionActionResult(BaseActionResult):
    """Result of getting wsproxy version."""

    wsproxy_version: str

    @override
    def entity_id(self) -> str | None:
        return None



"""BatchPurgerSpec implementations for container registry group associations."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import override

import sqlalchemy as sa

from ai.backend.manager.models.association_container_registries_groups import (
    AssociationContainerRegistriesGroupsRow,
)
from ai.backend.manager.repositories.base.purger import BatchPurgerSpec


@dataclass
class ContainerRegistryGroupPurgerSpec(
    BatchPurgerSpec[AssociationContainerRegistriesGroupsRow],
):
    """PurgerSpec for removing a container registry association from a project."""

    registry_id: uuid.UUID
    group_id: uuid.UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[AssociationContainerRegistriesGroupsRow]]:
        return sa.select(AssociationContainerRegistriesGroupsRow).where(
            sa.and_(
                AssociationContainerRegistriesGroupsRow.registry_id == self.registry_id,
                AssociationContainerRegistriesGroupsRow.group_id == self.group_id,
            )
        )



from __future__ import annotations

from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.actions.validators import ActionValidators

from .actions.fetch_status import FetchManagerStatusAction, FetchManagerStatusActionResult
from .actions.get_announcement import GetAnnouncementAction, GetAnnouncementActionResult
from .actions.get_db_cxn_status import GetDbCxnStatusAction, GetDbCxnStatusActionResult
from .actions.perform_scheduler_ops import (
    PerformSchedulerOpsAction,
    PerformSchedulerOpsActionResult,
)
from .actions.update_announcement import UpdateAnnouncementAction, UpdateAnnouncementActionResult
from .actions.update_status import UpdateManagerStatusAction, UpdateManagerStatusActionResult
from .service import ManagerAdminService

__all__ = ("ManagerAdminProcessors",)


class ManagerAdminProcessors(AbstractProcessorPackage):
    """Processor package for manager admin operations."""

    fetch_status: ActionProcessor[FetchManagerStatusAction, FetchManagerStatusActionResult]
    update_status: ActionProcessor[UpdateManagerStatusAction, UpdateManagerStatusActionResult]
    get_announcement: ActionProcessor[GetAnnouncementAction, GetAnnouncementActionResult]
    update_announcement: ActionProcessor[UpdateAnnouncementAction, UpdateAnnouncementActionResult]
    perform_scheduler_ops: ActionProcessor[
        PerformSchedulerOpsAction, PerformSchedulerOpsActionResult
    ]
    get_db_cxn_status: ActionProcessor[GetDbCxnStatusAction, GetDbCxnStatusActionResult]

    def __init__(
        self,
        service: ManagerAdminService,
        action_monitors: list[ActionMonitor],
        validators: ActionValidators,
    ) -> None:
        self.fetch_status = ActionProcessor(service.fetch_status, action_monitors)
        self.update_status = ActionProcessor(service.update_status, action_monitors)
        self.get_announcement = ActionProcessor(service.get_announcement, action_monitors)
        self.update_announcement = ActionProcessor(service.update_announcement, action_monitors)
        self.perform_scheduler_ops = ActionProcessor(service.perform_scheduler_ops, action_monitors)
        self.get_db_cxn_status = ActionProcessor(service.get_db_cxn_status, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            FetchManagerStatusAction.spec(),
            UpdateManagerStatusAction.spec(),
            GetAnnouncementAction.spec(),
            UpdateAnnouncementAction.spec(),
            PerformSchedulerOpsAction.spec(),
            GetDbCxnStatusAction.spec(),
        ]



from collections.abc import Awaitable, Callable
from typing import override

from ai.backend.common.types import KernelId
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.stream.actions.execute_in_stream import (
    ExecuteInStreamAction,
    ExecuteInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.gc_stale_connections import (
    GCStaleConnectionsAction,
    GCStaleConnectionsActionResult,
)
from ai.backend.manager.services.stream.actions.get_streaming_session import (
    GetStreamingSessionAction,
    GetStreamingSessionActionResult,
)
from ai.backend.manager.services.stream.actions.interrupt_in_stream import (
    InterruptInStreamAction,
    InterruptInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.restart_in_stream import (
    RestartInStreamAction,
    RestartInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.start_service_in_stream import (
    StartServiceInStreamAction,
    StartServiceInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.track_connection import (
    TrackConnectionAction,
    TrackConnectionActionResult,
)
from ai.backend.manager.services.stream.actions.untrack_connection import (
    UntrackConnectionAction,
    UntrackConnectionActionResult,
)
from ai.backend.manager.services.stream.service import StreamService


class StreamProcessors(AbstractProcessorPackage):
    _service: StreamService

    get_streaming_session: ActionProcessor[
        GetStreamingSessionAction, GetStreamingSessionActionResult
    ]
    track_connection: ActionProcessor[TrackConnectionAction, TrackConnectionActionResult]
    untrack_connection: ActionProcessor[UntrackConnectionAction, UntrackConnectionActionResult]
    gc_stale_connections: ActionProcessor[GCStaleConnectionsAction, GCStaleConnectionsActionResult]
    execute_in_stream: ActionProcessor[ExecuteInStreamAction, ExecuteInStreamActionResult]
    restart_in_stream: ActionProcessor[RestartInStreamAction, RestartInStreamActionResult]
    interrupt_in_stream: ActionProcessor[InterruptInStreamAction, InterruptInStreamActionResult]
    start_service_in_stream: ActionProcessor[
        StartServiceInStreamAction, StartServiceInStreamActionResult
    ]

    def __init__(self, service: StreamService, action_monitors: list[ActionMonitor]) -> None:
        self._service = service
        self.get_streaming_session = ActionProcessor(service.get_streaming_session, action_monitors)
        self.track_connection = ActionProcessor(service.track_connection, action_monitors)
        self.untrack_connection = ActionProcessor(service.untrack_connection, action_monitors)
        self.gc_stale_connections = ActionProcessor(service.gc_stale_connections, action_monitors)
        self.execute_in_stream = ActionProcessor(service.execute_in_stream, action_monitors)
        self.restart_in_stream = ActionProcessor(service.restart_in_stream, action_monitors)
        self.interrupt_in_stream = ActionProcessor(service.interrupt_in_stream, action_monitors)
        self.start_service_in_stream = ActionProcessor(
            service.start_service_in_stream, action_monitors
        )

    def create_connection_refresh_callback(
        self,
        kernel_id: KernelId,
        service: str,
        stream_id: str,
    ) -> Callable[..., Awaitable[None]]:
        return self._service.create_connection_refresh_callback(kernel_id, service, stream_id)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            GetStreamingSessionAction.spec(),
            TrackConnectionAction.spec(),
            UntrackConnectionAction.spec(),
            GCStaleConnectionsAction.spec(),
            ExecuteInStreamAction.spec(),
            RestartInStreamAction.spec(),
            InterruptInStreamAction.spec(),
            StartServiceInStreamAction.spec(),
        ]



"""Health module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import HealthHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_health_routes(
    handler: HealthHandler,
    route_deps: RouteDeps,
) -> RouteRegistry:
    """Build the health sub-application."""
    reg = RouteRegistry.create("health", route_deps.cors_options)

    # Public endpoint — no auth required
    reg.add("GET", "", handler.hello)

    return reg



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.permission import PermissionData
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.permission_contoller.actions.permission import PermissionAction


@dataclass
class UpdatePermissionAction(PermissionAction):
    updater: Updater[PermissionRow]

    @override
    def entity_id(self) -> str | None:
        return str(self.updater.pk_value)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpdatePermissionActionResult(BaseActionResult):
    data: PermissionData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id) if self.data else None



from abc import abstractmethod
from typing import TypeVar, override

from ai.backend.common.data.permission.types import ScopeType
from ai.backend.manager.data.permission.types import RBACElementRef

from .base import BaseAction, BaseActionResult


class BaseScopeAction(BaseAction):
    @override
    def entity_id(self) -> str | None:
        return None

    @abstractmethod
    def scope_type(self) -> ScopeType:
        raise NotImplementedError

    @abstractmethod
    def scope_id(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def target_element(self) -> RBACElementRef:
        """Return the RBAC element reference for the scope this action targets.

        Used by RBAC validators to check whether the current user has the
        required permission on this scope.

        Implementations must construct the RBACElementRef directly from their
        own fields — do not delegate to ``scope_type()`` / ``scope_id()``.
        """
        raise NotImplementedError


class BaseScopeActionResult(BaseActionResult):
    @override
    def entity_id(self) -> str | None:
        return None

    @abstractmethod
    def scope_type(self) -> ScopeType:
        raise NotImplementedError

    @abstractmethod
    def scope_id(self) -> str:
        raise NotImplementedError


TScopeAction = TypeVar("TScopeAction", bound=BaseScopeAction)
TScopeActionResult = TypeVar("TScopeActionResult", bound=BaseScopeActionResult)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.common.events.hub.hub import EventHub


class EventHubDependency(NonMonitorableDependencyProvider[None, EventHub]):
    """Provides EventHub lifecycle management."""

    @property
    def stage_name(self) -> str:
        return "event-hub"

    @asynccontextmanager
    async def provide(self, setup_input: None) -> AsyncIterator[EventHub]:
        """Initialize and provide the event hub.

        Args:
            setup_input: None (EventHub has no dependencies)

        Yields:
            Initialized EventHub
        """
        event_hub = EventHub()
        try:
            yield event_hub
        finally:
            await event_hub.shutdown()



from __future__ import annotations

import uuid

import sqlalchemy as sa

from ai.backend.common.data.filter_specs import StringMatchSpec
from ai.backend.manager.models.resource_slot import (
    AgentResourceRow,
    ResourceAllocationRow,
    ResourceSlotTypeRow,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class QueryConditions:
    @staticmethod
    def by_slot_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.slot_name.ilike(f"%{spec.value}%")
            else:
                condition = ResourceSlotTypeRow.slot_name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ResourceSlotTypeRow.slot_name) == spec.value.lower()
            else:
                condition = ResourceSlotTypeRow.slot_name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.slot_name.ilike(f"{spec.value}%")
            else:
                condition = ResourceSlotTypeRow.slot_name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.slot_name.ilike(f"%{spec.value}")
            else:
                condition = ResourceSlotTypeRow.slot_name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_type_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.slot_type.ilike(f"%{spec.value}%")
            else:
                condition = ResourceSlotTypeRow.slot_type.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_type_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ResourceSlotTypeRow.slot_type) == spec.value.lower()
            else:
                condition = ResourceSlotTypeRow.slot_type == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_type_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.slot_type.ilike(f"{spec.value}%")
            else:
                condition = ResourceSlotTypeRow.slot_type.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_type_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.slot_type.ilike(f"%{spec.value}")
            else:
                condition = ResourceSlotTypeRow.slot_type.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_display_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.display_name.ilike(f"%{spec.value}%")
            else:
                condition = ResourceSlotTypeRow.display_name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_display_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ResourceSlotTypeRow.display_name) == spec.value.lower()
            else:
                condition = ResourceSlotTypeRow.display_name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_display_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.display_name.ilike(f"{spec.value}%")
            else:
                condition = ResourceSlotTypeRow.display_name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_display_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceSlotTypeRow.display_name.ilike(f"%{spec.value}")
            else:
                condition = ResourceSlotTypeRow.display_name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner


class QueryOrders:
    @staticmethod
    def slot_name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ResourceSlotTypeRow.slot_name.asc()
        return ResourceSlotTypeRow.slot_name.desc()

    @staticmethod
    def rank(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ResourceSlotTypeRow.rank.asc()
        return ResourceSlotTypeRow.rank.desc()

    @staticmethod
    def display_name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ResourceSlotTypeRow.display_name.asc()
        return ResourceSlotTypeRow.display_name.desc()


class CursorConditions:
    @staticmethod
    def by_cursor_forward(cursor_slot_name: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor). slot_name is the primary key."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ResourceSlotTypeRow.slot_name > cursor_slot_name

        return inner

    @staticmethod
    def by_cursor_backward(cursor_slot_name: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor). slot_name is the primary key."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ResourceSlotTypeRow.slot_name < cursor_slot_name

        return inner


class AgentResourceQueryConditions:
    @staticmethod
    def by_agent_id(agent_id: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return AgentResourceRow.agent_id == agent_id

        return inner

    @staticmethod
    def by_slot_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = AgentResourceRow.slot_name.ilike(f"%{spec.value}%")
            else:
                condition = AgentResourceRow.slot_name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(AgentResourceRow.slot_name) == spec.value.lower()
            else:
                condition = AgentResourceRow.slot_name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = AgentResourceRow.slot_name.ilike(f"{spec.value}%")
            else:
                condition = AgentResourceRow.slot_name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = AgentResourceRow.slot_name.ilike(f"%{spec.value}")
            else:
                condition = AgentResourceRow.slot_name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_cursor_forward(cursor_slot_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return AgentResourceRow.slot_name > cursor_slot_name

        return inner

    @staticmethod
    def by_cursor_backward(cursor_slot_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return AgentResourceRow.slot_name < cursor_slot_name

        return inner


class AgentResourceQueryOrders:
    @staticmethod
    def slot_name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return AgentResourceRow.slot_name.asc()
        return AgentResourceRow.slot_name.desc()

    @staticmethod
    def capacity(ascending: bool = True) -> QueryOrder:
        if ascending:
            return AgentResourceRow.capacity.asc()
        return AgentResourceRow.capacity.desc()

    @staticmethod
    def used(ascending: bool = True) -> QueryOrder:
        if ascending:
            return AgentResourceRow.used.asc()
        return AgentResourceRow.used.desc()


class ResourceAllocationQueryConditions:
    @staticmethod
    def by_kernel_id(kernel_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ResourceAllocationRow.kernel_id == kernel_id

        return inner

    @staticmethod
    def by_slot_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceAllocationRow.slot_name.ilike(f"%{spec.value}%")
            else:
                condition = ResourceAllocationRow.slot_name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ResourceAllocationRow.slot_name) == spec.value.lower()
            else:
                condition = ResourceAllocationRow.slot_name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceAllocationRow.slot_name.ilike(f"{spec.value}%")
            else:
                condition = ResourceAllocationRow.slot_name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_slot_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ResourceAllocationRow.slot_name.ilike(f"%{spec.value}")
            else:
                condition = ResourceAllocationRow.slot_name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_cursor_forward(cursor_slot_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ResourceAllocationRow.slot_name > cursor_slot_name

        return inner

    @staticmethod
    def by_cursor_backward(cursor_slot_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ResourceAllocationRow.slot_name < cursor_slot_name

        return inner


class ResourceAllocationQueryOrders:
    @staticmethod
    def slot_name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ResourceAllocationRow.slot_name.asc()
        return ResourceAllocationRow.slot_name.desc()

    @staticmethod
    def requested(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ResourceAllocationRow.requested.asc()
        return ResourceAllocationRow.requested.desc()

    @staticmethod
    def used(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ResourceAllocationRow.used.asc()
        return ResourceAllocationRow.used.desc()



from __future__ import annotations

import functools
import inspect
import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Final

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import APIStreamResponse, extract_param_value, parse_response
from ai.backend.logging import BraceStyleAdapter

from .types import ApiHandler, CORSOptions, RouteMiddleware, WebRequestHandler

if TYPE_CHECKING:
    from aiohttp.typedefs import Middleware

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


async def _handle_stream_response(
    request: web.Request,
    result: APIStreamResponse,
) -> web.StreamResponse:
    """Convert an ``APIStreamResponse`` into a ``web.StreamResponse`` with chunked streaming."""
    resp = web.StreamResponse(status=result.status, headers=result.headers)
    body_iter = result.body.read()

    try:
        first_chunk = await body_iter.__anext__()
        await resp.prepare(request)
        await resp.write(first_chunk)
    except Exception:
        log.exception("Failed to send first chunk from stream")
        raise web.HTTPInternalServerError(
            reason="Failed to initialize streaming response"
        ) from None

    try:
        async for chunk in body_iter:
            await resp.write(chunk)
        await resp.write_eof()
    except Exception:
        log.exception("Error during streaming response body iteration")
        resp.force_close()

    return resp


def _wrap_api_handler(handler: ApiHandler) -> WebRequestHandler:
    """Wrap a typed API handler: extract params from web.Request, convert APIResponse → web.Response."""
    sig = inspect.signature(handler, eval_str=True)

    @functools.wraps(handler)
    async def wrapped(request: web.Request) -> web.StreamResponse:
        kwargs: dict[str, Any] = {}
        for name, param in sig.parameters.items():
            kwargs[name] = await extract_param_value(request, param.annotation)
        response = await handler(**kwargs)
        if isinstance(response, web.StreamResponse):
            return response
        if isinstance(response, APIStreamResponse):
            return await _handle_stream_response(request, response)
        return parse_response(response)

    return wrapped


async def _on_prepare(_request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


class RouteRegistry:
    """Registry for API routes with parent-child tree support.

    Each registry owns a ``web.Application`` and a URL prefix.  Registries
    form a tree via ``add_subregistry()``.  At mount time the tree is
    flattened by ``collect_apps()`` so that every sub-application is
    mounted directly on the root application (no nested subapps).

    Middleware tiers (outermost → innermost):

    1. **Registry-level middleware** — applied to every route registered
       through this registry instance (includes middlewares inherited from
       parent registries via ``add_subregistry()``).
    2. **Route-level middleware** — applied to individual routes via the
       ``add()`` method (e.g. admin_required on a specific endpoint).

    When both registry-level and route-level middlewares are present,
    they are concatenated (registry first, then route) and applied as a
    single chain in declaration order.
    """

    _prefix: str
    _app: web.Application
    _cors: aiohttp_cors.CorsConfig
    _middlewares: list[RouteMiddleware]
    _subregistries: dict[str, RouteRegistry]
    rlim_middleware: Middleware | None

    def __init__(
        self,
        app: web.Application,
        cors_options: CORSOptions,
        *,
        prefix: str = "",
        middlewares: Sequence[RouteMiddleware] | None = None,
    ) -> None:
        self._prefix = prefix
        self._app = app
        self._cors = aiohttp_cors.setup(app, defaults=cors_options)
        self._middlewares = list(middlewares or [])
        self._subregistries = {}
        self.rlim_middleware = None

    @classmethod
    def create(
        cls,
        prefix: str,
        cors_options: CORSOptions,
        *,
        middlewares: Sequence[RouteMiddleware] = (),
    ) -> RouteRegistry:
        """Create a new sub-application with its own ``RouteRegistry``.

        The sub-application is not yet mounted; call ``add_subregistry()``
        on a parent registry to wire it into the tree, then use
        ``collect_apps()`` at mount time.
        """
        subapp = web.Application()
        subapp.on_response_prepare.append(_on_prepare)
        return cls(subapp, cors_options, prefix=prefix, middlewares=middlewares)

    @property
    def prefix(self) -> str:
        return self._prefix

    @property
    def app(self) -> web.Application:
        return self._app

    @property
    def cors(self) -> aiohttp_cors.CorsConfig:
        return self._cors

    @property
    def middlewares(self) -> list[RouteMiddleware]:
        return self._middlewares

    def add_subregistry(self, child: RouteRegistry) -> None:
        """Register a child registry in the tree.

        Aggregates ``self.prefix + child.prefix`` into the child's stored
        prefix so that ``collect_apps()`` returns fully-qualified prefixes.
        Propagates parent middlewares into the child.
        Raises ``ValueError`` on duplicate prefix.
        """
        child_prefix = child.prefix
        if child_prefix in self._subregistries:
            raise ValueError(f"Sub-registry with prefix '{child_prefix}' already registered")
        # Propagate parent middlewares into child (parent outermost)
        child._middlewares = self._middlewares + child._middlewares
        # Aggregate prefix: "admin" + "domains" → "admin/domains"
        if self._prefix:
            child._prefix = f"{self._prefix}/{child._prefix}"
        self._subregistries[child_prefix] = child

    def find_subregistry(self, prefix: str) -> RouteRegistry | None:
        """Look up a direct sub-registry by its original (non-aggregated) prefix."""
        return self._subregistries.get(prefix)

    def collect_apps(self) -> list[tuple[str, web.Application, RouteRegistry]]:
        """Flatten the tree into ``(aggregated_prefix, app, registry)`` tuples.

        Example: admin registry (prefix="admin") with domain sub-registry
        (prefix="domains") yields::

            [("admin", admin_app, admin_reg),
             ("admin/domains", domain_app, domain_reg)]

        Leaf registries without routes are included only if they have no
        sub-registries (i.e. they are true leaves).
        """
        result: list[tuple[str, web.Application, RouteRegistry]] = []
        if list(self._app.router.routes()) or not self._subregistries:
            result.append((self._prefix, self._app, self))
        for child in self._subregistries.values():
            result.extend(child.collect_apps())
        return result

    def add(
        self,
        method: str,
        path: str,
        handler: ApiHandler,
        middlewares: Sequence[RouteMiddleware] | None = None,
    ) -> web.AbstractRoute:
        """Register a route with optional per-route middleware.

        The handler is always wrapped by ``_wrap_api_handler`` so that
        typed parameters (``BodyParam``, ``QueryParam``, ``MiddlewareParam``,
        etc.) are automatically extracted from the ``web.Request`` and the
        returned ``APIResponse`` is converted to a ``web.Response``.

        Registry-level middlewares (including those inherited from parent
        registries) are prepended to the per-route list, so they wrap the
        handler outermost.

        Example::

            registry = RouteRegistry.create("admin", cors, middlewares=[auth_required])
            registry.add("GET", "/users", list_users, middlewares=[
                admin_required,
            ])

        is equivalent to::

            @auth_required      # registry-level (outermost)
            @admin_required     # route-level
            async def list_users(request): ...
        """
        effective_handler: WebRequestHandler = _wrap_api_handler(handler)
        combined = self._middlewares + list(middlewares or [])
        final_handler: WebRequestHandler = effective_handler
        if combined:
            final_handler = _apply_route_middlewares(effective_handler, combined)
        route = self._app.router.add_route(method, path, final_handler)
        self._cors.add(route)
        return route


def _apply_route_middlewares(
    handler: WebRequestHandler,
    middlewares: Sequence[RouteMiddleware],
) -> WebRequestHandler:
    """Wrap a handler with a chain of route-level middlewares.

    Middlewares are applied so that the first element in the list
    is the outermost wrapper, preserving standard decorator order.
    """
    wrapped: WebRequestHandler = handler
    for middleware in reversed(middlewares):
        wrapped = middleware(wrapped)
    return wrapped



from __future__ import annotations

from typing import TYPE_CHECKING

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.models.agent import AgentRow, AgentStatus

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("EtcdConfigDBSource",)

etcd_config_db_source_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.DB_SOURCE, layer=LayerType.ETCD_CONFIG_DB_SOURCE)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=5,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class EtcdConfigDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @etcd_config_db_source_resilience.apply()
    async def get_available_agent_slots(self, sgroup: str) -> set[str]:
        """Get the set of available slot keys from alive, schedulable agents in a scaling group."""
        available_slot_keys: set[str] = set()
        async with self._db.begin_readonly_session() as db_sess:
            result = await db_sess.execute(
                sa.select(AgentRow).where(
                    (AgentRow.status == AgentStatus.ALIVE)
                    & (AgentRow.scaling_group == sgroup)
                    & (AgentRow.schedulable == sa.true())
                )
            )
            for agent in result.scalars().all():
                available_slot_keys.update(agent.available_slots.keys())
        return available_slot_keys



from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.role import (
    BulkRoleAssignmentResultData,
)
from ai.backend.manager.models.rbac_models.user_role import UserRoleRow
from ai.backend.manager.repositories.base.creator import BulkCreator
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class BulkAssignRoleAction(RoleAction):
    bulk_creator: BulkCreator[UserRoleRow]

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.ROLE_ASSIGNMENT

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class BulkAssignRoleActionResult(BaseActionResult):
    data: BulkRoleAssignmentResultData

    @override
    def entity_id(self) -> str | None:
        return None



"""Prometheus query preset GQL Node, Edge, Connection, and mutation payload types."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Self

import strawberry
from strawberry import ID
from strawberry.relay import Connection, Edge, Node, NodeID

from .payloads import QueryDefinitionOptionsGQL

if TYPE_CHECKING:
    from ai.backend.manager.data.prometheus_query_preset import PrometheusQueryPresetData


@strawberry.type(
    name="QueryDefinition",
    description="Added in 26.3.0. Prometheus query definition entity implementing Relay Node pattern.",
)
class QueryDefinitionGQL(Node):
    id: NodeID[str] = strawberry.field(description="Query definition UUID (primary key).")
    name: str = strawberry.field(description="Human-readable query definition identifier.")
    metric_name: str = strawberry.field(description="Prometheus metric name.")
    query_template: str = strawberry.field(description="PromQL template with placeholders.")
    time_window: str | None = strawberry.field(
        description="Default time window. Falls back to server config if null."
    )
    options: QueryDefinitionOptionsGQL = strawberry.field(
        description="Query definition options including filter and group labels."
    )
    created_at: datetime = strawberry.field(description="Creation timestamp.")
    updated_at: datetime = strawberry.field(description="Last update timestamp.")

    @classmethod
    def from_data(cls, data: PrometheusQueryPresetData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            metric_name=data.metric_name,
            query_template=data.query_template,
            time_window=data.time_window,
            options=QueryDefinitionOptionsGQL(
                filter_labels=data.filter_labels,
                group_labels=data.group_labels,
            ),
            created_at=data.created_at,
            updated_at=data.updated_at,
        )


QueryDefinitionEdge = Edge[QueryDefinitionGQL]


@strawberry.type(
    description="Added in 26.3.0. Paginated connection for query definition records.",
)
class QueryDefinitionConnection(Connection[QueryDefinitionGQL]):
    count: int = strawberry.field(
        description="Total number of query definition records matching the query criteria."
    )

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.type(
    name="CreateQueryDefinitionPayload",
    description="Added in 26.3.0. Payload returned after creating a query definition.",
)
class CreateQueryDefinitionPayload:
    preset: QueryDefinitionGQL


@strawberry.type(
    name="ModifyQueryDefinitionPayload",
    description="Added in 26.3.0. Payload returned after modifying a query definition.",
)
class ModifyQueryDefinitionPayload:
    preset: QueryDefinitionGQL



"""Prometheus query preset GQL filter and order-by types."""

from __future__ import annotations

from enum import StrEnum
from typing import override

import strawberry

from ai.backend.manager.api.gql.base import (
    OrderDirection,
    StringFilter,
)
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.prometheus_query_preset.options import (
    PrometheusQueryPresetConditions,
    PrometheusQueryPresetOrders,
)


@strawberry.input(
    name="QueryDefinitionFilter",
    description="Added in 26.3.0. Filter input for querying query definitions.",
)
class QueryDefinitionFilter(GQLFilter):
    name: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by name.",
    )

    AND: list[QueryDefinitionFilter] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with AND logic.",
    )
    OR: list[QueryDefinitionFilter] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with OR logic.",
    )
    NOT: list[QueryDefinitionFilter] | None = strawberry.field(
        default=None,
        description="Negate the specified filters.",
    )

    @override
    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []

        if self.name:
            condition = self.name.build_query_condition(
                contains_factory=lambda spec: PrometheusQueryPresetConditions.by_name_contains(
                    spec
                ),
                equals_factory=lambda spec: PrometheusQueryPresetConditions.by_name_equals(spec),
                starts_with_factory=lambda spec: PrometheusQueryPresetConditions.by_name_starts_with(
                    spec
                ),
                ends_with_factory=lambda spec: PrometheusQueryPresetConditions.by_name_ends_with(
                    spec
                ),
            )
            if condition:
                conditions.append(condition)

        if self.AND:
            for sub_filter in self.AND:
                conditions.extend(sub_filter.build_conditions())

        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                conditions.append(combine_conditions_or(or_sub_conditions))

        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                conditions.append(negate_conditions(not_sub_conditions))

        return conditions


@strawberry.enum(
    name="QueryDefinitionOrderField",
    description="Added in 26.3.0. Fields available for ordering query definition results.",
)
class QueryDefinitionOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"
    NAME = "name"


@strawberry.input(
    name="QueryDefinitionOrderBy",
    description="Added in 26.3.0. Specifies ordering for query definition results.",
)
class QueryDefinitionOrderBy(GQLOrderBy):
    field: QueryDefinitionOrderField = strawberry.field(description="The field to order by.")
    direction: OrderDirection = strawberry.field(
        default=OrderDirection.DESC,
        description="Sort direction.",
    )

    @override
    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case QueryDefinitionOrderField.CREATED_AT:
                return PrometheusQueryPresetOrders.created_at(ascending)
            case QueryDefinitionOrderField.UPDATED_AT:
                return PrometheusQueryPresetOrders.updated_at(ascending)
            case QueryDefinitionOrderField.NAME:
                return PrometheusQueryPresetOrders.name(ascending)



"""GraphQL resolvers for RBAC role management."""

from __future__ import annotations

import uuid

import strawberry
from strawberry import ID, Info

from ai.backend.common.contexts.user import current_user
from ai.backend.common.data.permission.types import RoleStatus
from ai.backend.manager.api.gql.rbac.fetcher.role import (
    fetch_role,
    fetch_role_assignments,
    fetch_roles,
)
from ai.backend.manager.api.gql.rbac.types import (
    AssignRoleInput,
    BulkAssignRoleErrorGQL,
    BulkAssignRoleInputGQL,
    BulkAssignRolePayloadGQL,
    BulkRevokeRoleErrorGQL,
    BulkRevokeRoleInputGQL,
    BulkRevokeRolePayloadGQL,
    CreateRoleInput,
    DeleteRoleInput,
    DeleteRolePayload,
    PurgeRoleInput,
    PurgeRolePayload,
    RevokeRoleInput,
    RoleAssignmentConnection,
    RoleAssignmentFilter,
    RoleAssignmentGQL,
    RoleAssignmentOrderBy,
    RoleConnection,
    RoleFilter,
    RoleGQL,
    RoleOrderBy,
    UpdateRoleInput,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.repositories.base.purger import Purger
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.permission_controller.options import AssignedUserConditions
from ai.backend.manager.repositories.permission_controller.updaters import RoleUpdaterSpec
from ai.backend.manager.services.permission_contoller.actions.assign_role import (
    AssignRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.bulk_assign_role import (
    BulkAssignRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.bulk_revoke_role import (
    BulkRevokeRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.create_role import (
    CreateRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.delete_role import (
    DeleteRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.purge_role import (
    PurgeRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.revoke_role import (
    RevokeRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.update_role import (
    UpdateRoleAction,
)
from ai.backend.manager.types import OptionalState

# ==================== Query Resolvers ====================


@strawberry.field(description="Added in 26.3.0. Get a single role by ID (admin only).")  # type: ignore[misc]
async def admin_role(
    info: Info[StrawberryGQLContext],
    id: uuid.UUID,
) -> RoleGQL | None:
    check_admin_only()
    return await fetch_role(info, id)


@strawberry.field(
    description="Added in 26.3.0. List roles with filtering and pagination (admin only)."
)  # type: ignore[misc]
async def admin_roles(
    info: Info[StrawberryGQLContext],
    filter: RoleFilter | None = None,
    order_by: list[RoleOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> RoleConnection:
    check_admin_only()
    return await fetch_roles(
        info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description="Added in 26.3.0. List role assignments with filtering and pagination (admin only)."
)  # type: ignore[misc]
async def admin_role_assignments(
    info: Info[StrawberryGQLContext],
    filter: RoleAssignmentFilter | None = None,
    order_by: list[RoleAssignmentOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> RoleAssignmentConnection:
    check_admin_only()
    return await fetch_role_assignments(
        info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description="Added in 26.3.0. List roles assigned to the current authenticated user.",
)  # type: ignore[misc]
async def my_roles(
    info: Info[StrawberryGQLContext],
    filter: RoleAssignmentFilter | None = None,
    order_by: list[RoleAssignmentOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> RoleAssignmentConnection:
    me = current_user()
    if me is None:
        from ai.backend.manager.errors.auth import InsufficientPrivilege

        raise InsufficientPrivilege("Authentication required")

    return await fetch_role_assignments(
        info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
        base_conditions=[AssignedUserConditions.by_user_id(me.user_id)],
    )


# ==================== Mutation Resolvers ====================


@strawberry.mutation(description="Added in 26.3.0. Create a new role (admin only).")  # type: ignore[misc]
async def admin_create_role(
    info: Info[StrawberryGQLContext],
    input: CreateRoleInput,
) -> RoleGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.create_role.wait_for_complete(
            CreateRoleAction(creator=input.to_creator())
        )
    )
    return RoleGQL.from_dataclass(action_result.data)


@strawberry.mutation(description="Added in 26.3.0. Update an existing role (admin only).")  # type: ignore[misc]
async def admin_update_role(
    info: Info[StrawberryGQLContext],
    input: UpdateRoleInput,
) -> RoleGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.update_role.wait_for_complete(
            UpdateRoleAction(updater=input.to_updater())
        )
    )
    return RoleGQL.from_dataclass(action_result.data)


@strawberry.mutation(description="Added in 26.3.0. Soft-delete a role (admin only).")  # type: ignore[misc]
async def admin_delete_role(
    info: Info[StrawberryGQLContext],
    input: DeleteRoleInput,
) -> DeleteRolePayload:
    check_admin_only()
    updater = Updater(
        spec=RoleUpdaterSpec(
            status=OptionalState.update(RoleStatus.DELETED),
        ),
        pk_value=input.id,
    )
    await info.context.processors.permission_controller.delete_role.wait_for_complete(
        DeleteRoleAction(updater=updater)
    )
    return DeleteRolePayload(id=ID(str(input.id)))


@strawberry.mutation(description="Added in 26.3.0. Permanently remove a role (admin only).")  # type: ignore[misc]
async def admin_purge_role(
    info: Info[StrawberryGQLContext],
    input: PurgeRoleInput,
) -> PurgeRolePayload:
    check_admin_only()
    purger = Purger(row_class=RoleRow, pk_value=input.id)
    await info.context.processors.permission_controller.purge_role.wait_for_complete(
        PurgeRoleAction(purger=purger)
    )
    return PurgeRolePayload(id=ID(str(input.id)))


@strawberry.mutation(description="Added in 26.3.0. Assign a role to a user (admin only).")  # type: ignore[misc]
async def admin_assign_role(
    info: Info[StrawberryGQLContext],
    input: AssignRoleInput,
) -> RoleAssignmentGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.assign_role.wait_for_complete(
            AssignRoleAction(input=input.to_input())
        )
    )
    return RoleAssignmentGQL.from_assignment_data(action_result.data)


@strawberry.mutation(description="Added in 26.3.0. Revoke a role from a user (admin only).")  # type: ignore[misc]
async def admin_revoke_role(
    info: Info[StrawberryGQLContext],
    input: RevokeRoleInput,
) -> RoleAssignmentGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.revoke_role.wait_for_complete(
            RevokeRoleAction(input=input.to_input())
        )
    )
    return RoleAssignmentGQL.from_revocation_data(action_result.data)


@strawberry.mutation(
    description="Added in 26.3.0. Bulk assign a role to multiple users (admin only)."
)  # type: ignore[misc]
async def admin_bulk_assign_role(
    info: Info[StrawberryGQLContext],
    input: BulkAssignRoleInputGQL,
) -> BulkAssignRolePayloadGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.bulk_assign_role.wait_for_complete(
            BulkAssignRoleAction(bulk_creator=input.to_bulk_creator())
        )
    )
    return BulkAssignRolePayloadGQL(
        assigned=[RoleAssignmentGQL.from_assignment_data(s) for s in action_result.data.successes],
        failed=[
            BulkAssignRoleErrorGQL(user_id=f.user_id, message=f.message)
            for f in action_result.data.failures
        ],
    )


@strawberry.mutation(
    description="Added in 26.3.0. Bulk revoke a role from multiple users (admin only)."
)  # type: ignore[misc]
async def admin_bulk_revoke_role(
    info: Info[StrawberryGQLContext],
    input: BulkRevokeRoleInputGQL,
) -> BulkRevokeRolePayloadGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.bulk_revoke_role.wait_for_complete(
            BulkRevokeRoleAction(input=input.to_input())
        )
    )
    return BulkRevokeRolePayloadGQL(
        revoked=[RoleAssignmentGQL.from_revocation_data(s) for s in action_result.data.successes],
        failed=[
            BulkRevokeRoleErrorGQL(user_id=f.user_id, message=f.message)
            for f in action_result.data.failures
        ],
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import ContainerRegistryAction


@dataclass
class HarborWebhookResourceInput:
    """A single resource from a Harbor webhook payload."""

    resource_url: str
    tag: str


@dataclass
class HandleHarborWebhookAction(ContainerRegistryAction):
    """Action to handle a Harbor container registry webhook event."""

    event_type: str
    resources: list[HarborWebhookResourceInput] = field(default_factory=list)
    project: str = ""
    img_name: str = ""
    auth_header: str | None = None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class HandleHarborWebhookActionResult(BaseActionResult):
    """Result of handling a Harbor webhook event."""

    @override
    def entity_id(self) -> str | None:
        return None



from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action.scope import BaseScopeAction, BaseScopeActionResult


class AppConfigScopeAction(BaseScopeAction):
    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG


class AppConfigScopeActionResult(BaseScopeActionResult):
    pass



from __future__ import annotations

import asyncio
import logging
import traceback
from typing import TYPE_CHECKING, Final, cast

from aiohttp import web
from aiohttp.typedefs import Middleware

from ai.backend.common.exception import BackendAIError, ErrorCode
from ai.backend.common.json import dump_json_str
from ai.backend.common.plugin.monitor import INCREMENT
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.errors.common import (
    GenericBadRequest,
    InternalServerError,
    MethodNotAllowed,
    URLNotFound,
)
from ai.backend.manager.exceptions import InvalidArgument

if TYPE_CHECKING:
    from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
    from ai.backend.manager.api.rest.types import WebRequestHandler
    from ai.backend.manager.config.provider import ManagerConfigProvider

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


def _debug_error_response(
    e: Exception,
) -> web.StreamResponse:
    error_type = ""
    error_title = ""
    error_message = "Internal server error"
    status_code = 500
    error_code = ErrorCode.default()
    if isinstance(e, BackendAIError):
        error_type = e.error_type
        error_title = e.error_title
        if e.extra_msg:
            error_message = e.extra_msg
        status_code = e.status_code
        error_code = e.error_code()

    return web.json_response(
        {
            "type": error_type,
            "title": error_title,
            "error_code": str(error_code),
            "msg": error_message,
            "traceback": traceback.format_exc(),
        },
        status=status_code,
        dumps=dump_json_str,
    )


def build_exception_middleware(
    *,
    error_monitor: ErrorPluginContext,
    stats_monitor: StatsPluginContext,
    config_provider: ManagerConfigProvider,
) -> Middleware:
    """Build an exception middleware with explicit dependencies."""

    @web.middleware
    async def _middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
        method = request.method
        endpoint = getattr(request.match_info.route.resource, "canonical", request.path)
        try:
            await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.requests")
            resp = await handler(request)
        except InvalidArgument as ex:
            if len(ex.args) > 1:
                raise InvalidAPIParameters(
                    f"{ex.args[0]}: {', '.join(map(str, ex.args[1:]))}"
                ) from ex
            if len(ex.args) == 1:
                raise InvalidAPIParameters(ex.args[0]) from ex
            raise InvalidAPIParameters() from ex
        except BackendAIError as ex:
            if ex.status_code // 100 == 4:
                log.warning(
                    "client error raised inside handlers: ({} {}): {}",
                    method,
                    endpoint,
                    repr(ex),
                )
            elif ex.status_code // 100 == 5:
                log.exception(
                    "Internal server error raised inside handlers: ({} {}): {}",
                    method,
                    endpoint,
                    repr(ex),
                )
            await error_monitor.capture_exception()
            await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
            await stats_monitor.report_metric(
                INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
            )
            if config_provider.config.debug.enabled:
                return _debug_error_response(ex)
            raise
        except web.HTTPException as ex:
            await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
            await stats_monitor.report_metric(
                INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
            )
            if ex.status_code // 100 == 4:
                log.warning(
                    "client error raised inside handlers: ({} {}): {}", method, endpoint, ex
                )
            elif ex.status_code // 100 == 5:
                log.exception(
                    "Internal server error raised inside handlers: ({} {}): {}",
                    method,
                    endpoint,
                    ex,
                )
            if ex.status_code == 404:
                raise URLNotFound(extra_data=request.path) from ex
            if ex.status_code == 405:
                concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
                raise MethodNotAllowed(
                    method=concrete_ex.method, allowed_methods=concrete_ex.allowed_methods
                ) from ex
            raise GenericBadRequest from ex
        except asyncio.CancelledError as e:
            log.debug("Request cancelled ({0} {1})", request.method, request.rel_url)
            raise e
        except Exception as e:
            await error_monitor.capture_exception()
            log.exception(
                "Uncaught exception in HTTP request handlers ({} {}): {}", method, endpoint, e
            )
            if config_provider.config.debug.enabled:
                return _debug_error_response(e)
            raise InternalServerError() from e
        else:
            await stats_monitor.report_metric(
                INCREMENT, f"ai.backend.manager.api.status.{resp.status}"
            )
            return resp

    return _middleware



"""Creator for RBAC scope-scoped entity insert operations."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TypeVar

import sqlalchemy as sa
from sqlalchemy import inspect
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.data.permission.types import RBACElementType, RelationType
from ai.backend.manager.data.permission.types import RBACElementRef
from ai.backend.manager.errors.repository import UnsupportedCompositePrimaryKeyError
from ai.backend.manager.models.base import Base
from ai.backend.manager.models.rbac_models.association_scopes_entities import (
    AssociationScopesEntitiesRow,
)
from ai.backend.manager.repositories.base.creator import CreatorSpec
from ai.backend.manager.repositories.base.integrity import (
    match_integrity_error,
    parse_integrity_error,
)

TRow = TypeVar("TRow", bound=Base)


# =============================================================================
# Single Entity Creator
# =============================================================================


@dataclass
class RBACEntityCreator[TRow: Base]:
    """Creator for a single entity with scope associations for RBAC.

    Creates an entity row and associates it with one or more permission scopes.
    The primary scope is required; additional scopes are optional.

    Attributes:
        spec: CreatorSpec implementation defining the row to create.
        element_type: The RBAC element type for this entity.
        scope_ref: Primary scope reference (scope_type + scope_id) for this entity.
        additional_scope_refs: Additional scope references for multi-scope entities.
        relation_type: The relation type for the scope-entity association. Defaults to AUTO.
    """

    spec: CreatorSpec[TRow]
    element_type: RBACElementType
    scope_ref: RBACElementRef
    additional_scope_refs: Sequence[RBACElementRef] = field(default_factory=list)
    relation_type: RelationType = RelationType.AUTO


@dataclass
class RBACEntityCreatorResult[TRow: Base]:
    """Result of executing a single entity creation."""

    row: TRow


async def execute_rbac_entity_creator[TRow: Base](
    db_sess: SASession,
    creator: RBACEntityCreator[TRow],
) -> RBACEntityCreatorResult[TRow]:
    """Create a scope-scoped entity with its scope association.

    Operations:
    1. Insert main entity row
    2. Flush to get DB-generated ID
    3. Extract RBAC info from spec
    4. Insert AssociationScopesEntitiesRow (scope -> entity mapping)

    The AssociationScopesEntitiesRow maps the entity to its owning scope,
    enabling scope-based entity discovery and permission inheritance.

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creator: Creator instance with spec defining the entity to create.

    Returns:
        RBACEntityCreatorResult containing the created row.
    """
    spec = creator.spec
    row = spec.build_row()
    mapper = inspect(type(row))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Entity creator only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 1. Build and insert row
    db_sess.add(row)

    # 2. Flush to get DB-generated ID
    try:
        await db_sess.flush()
    except sa.exc.IntegrityError as e:
        parsed = parse_integrity_error(e)
        match_integrity_error(parsed, spec.integrity_error_checks)

    # 3. Extract RBAC info and insert associations for all scopes
    instance_state = inspect(row)
    pk_value = instance_state.identity[0]
    entity_type = creator.element_type.to_entity_type()
    all_scope_refs = [creator.scope_ref, *creator.additional_scope_refs]
    for scope_ref in all_scope_refs:
        db_sess.add(
            AssociationScopesEntitiesRow(
                scope_type=scope_ref.element_type.to_scope_type(),
                scope_id=scope_ref.element_id,
                entity_type=entity_type,
                entity_id=str(pk_value),
                relation_type=creator.relation_type,
            ),
        )

    return RBACEntityCreatorResult(row=row)


# =============================================================================
# Bulk Entity Creator
# =============================================================================


@dataclass
class RBACBulkEntityCreator[TRow: Base]:
    """Bulk creator for multiple entities with a single shared scope.

    Attributes:
        specs: Sequence of CreatorSpec implementations.
        element_type: The RBAC element type for all entities.
        scope_ref: The scope reference (scope_type + scope_id) for all entities.
        relation_type: The relation type for the scope-entity association. Defaults to AUTO.
    """

    specs: Sequence[CreatorSpec[TRow]]
    element_type: RBACElementType
    scope_ref: RBACElementRef
    relation_type: RelationType = RelationType.AUTO


@dataclass
class RBACBulkEntityCreatorResult[TRow: Base]:
    """Result of executing a bulk entity creation."""

    rows: list[TRow]


async def execute_rbac_bulk_entity_creator[TRow: Base](
    db_sess: SASession,
    creator: RBACBulkEntityCreator[TRow],
) -> RBACBulkEntityCreatorResult[TRow]:
    """Create multiple scope-scoped entities in a single transaction.

    Operations:
    1. Build and insert all entity rows
    2. Flush to get DB-generated IDs
    3. Bulk insert AssociationScopesEntitiesRows

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creator: Bulk creator with specs defining entities to create.

    Returns:
        RBACBulkEntityCreatorResult containing all created rows.
    """
    if not creator.specs:
        return RBACBulkEntityCreatorResult(rows=[])

    # 1. Build and add all rows
    rows = [spec.build_row() for spec in creator.specs]
    db_sess.add_all(rows)

    mapper = inspect(type(rows[0]))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Entity creator only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 2. Flush to get DB-generated IDs and insert associations
    try:
        await db_sess.flush()
    except sa.exc.IntegrityError as e:
        parsed = parse_integrity_error(e)
        # Use first spec's checks (all specs share the same CreatorSpec subclass)
        checks = creator.specs[0].integrity_error_checks
        match_integrity_error(parsed, checks)

    entity_type = creator.element_type.to_entity_type()
    associations = [
        AssociationScopesEntitiesRow(
            scope_type=creator.scope_ref.element_type.to_scope_type(),
            scope_id=creator.scope_ref.element_id,
            entity_type=entity_type,
            entity_id=str(inspect(row).identity[0]),
            relation_type=creator.relation_type,
        )
        for row in rows
    ]
    db_sess.add_all(associations)

    return RBACBulkEntityCreatorResult(rows=rows)


async def execute_rbac_entity_creators[TRow: Base](
    db_sess: SASession,
    creators: Sequence[RBACEntityCreator[TRow]],
) -> RBACBulkEntityCreatorResult[TRow]:
    """Create multiple entities from individual RBACEntityCreator instances in a single batch.

    Unlike execute_rbac_bulk_entity_creator which shares a single scope across all entities,
    this function accepts a sequence of RBACEntityCreator instances where each entity can have
    its own primary scope and additional scope references.

    Operations:
    1. Build and insert all entity rows from each creator's spec
    2. Single flush to get all DB-generated IDs
    3. Collect all scope associations (primary + additional per entity), bulk insert
    4. Return RBACBulkEntityCreatorResult with all rows

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creators: Sequence of RBACEntityCreator instances, each with its own scope refs.

    Returns:
        RBACBulkEntityCreatorResult containing all created rows.
    """
    if not creators:
        return RBACBulkEntityCreatorResult(rows=[])

    # 1. Build and add all rows
    rows = [creator.spec.build_row() for creator in creators]
    db_sess.add_all(rows)

    mapper = inspect(type(rows[0]))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Entity creator only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 2. Single flush to get all DB-generated IDs
    try:
        await db_sess.flush()
    except sa.exc.IntegrityError as e:
        parsed = parse_integrity_error(e)
        # Use first creator's spec checks (all creators share the same CreatorSpec subclass)
        checks = creators[0].spec.integrity_error_checks
        match_integrity_error(parsed, checks)

    # 3. Collect all associations from each creator's scope refs
    associations: list[AssociationScopesEntitiesRow] = []
    for creator, row in zip(creators, rows, strict=True):
        pk_value = inspect(row).identity[0]
        entity_type = creator.element_type.to_entity_type()
        all_scope_refs = [creator.scope_ref, *creator.additional_scope_refs]
        for scope_ref in all_scope_refs:
            associations.append(
                AssociationScopesEntitiesRow(
                    scope_type=scope_ref.element_type.to_scope_type(),
                    scope_id=scope_ref.element_id,
                    entity_type=entity_type,
                    entity_id=str(pk_value),
                    relation_type=creator.relation_type,
                ),
            )
    db_sess.add_all(associations)

    return RBACBulkEntityCreatorResult(rows=rows)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING

from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.manager.notification.notification_center import NotificationCenter
from ai.backend.manager.repositories.repositories import Repositories
from ai.backend.manager.service.base import ServicesContext
from ai.backend.manager.types import DistributedLockFactory

from .distributed_lock import DistributedLockFactoryDependency, DistributedLockInput
from .notification import NotificationCenterDependency
from .repositories import RepositoriesDependency, RepositoriesInput
from .services import ServicesContextDependency, ServicesInput

if TYPE_CHECKING:
    from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
    from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
    from ai.backend.common.clients.valkey_client.valkey_schedule.client import (
        ValkeyScheduleClient,
    )
    from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
    from ai.backend.common.etcd import AsyncEtcd
    from ai.backend.manager.config.provider import ManagerConfigProvider
    from ai.backend.manager.models.storage import StorageSessionManager
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


@dataclass
class DomainInput:
    """Input required for domain setup.

    Contains infrastructure and component resources from earlier stages.
    """

    config_provider: ManagerConfigProvider
    db: ExtendedAsyncSAEngine
    etcd: AsyncEtcd
    storage_manager: StorageSessionManager
    valkey_stat: ValkeyStatClient
    valkey_live: ValkeyLiveClient
    valkey_schedule: ValkeyScheduleClient
    valkey_image: ValkeyImageClient


@dataclass
class DomainResources:
    """Container for all domain resources."""

    notification_center: NotificationCenter
    distributed_lock_factory: DistributedLockFactory
    repositories: Repositories
    services_ctx: ServicesContext


class DomainComposer(DependencyComposer[DomainInput, DomainResources]):
    """Composes all domain dependencies.

    Composes repositories and domain objects at Layer 0+3:
    1. Notification center: HTTP client pool for notifications (no deps)
    2. Distributed lock factory: Lock backend based on config
    3. Repositories: All repository instances
    4. Services context: Service-layer objects for the API layer
    """

    @property
    def stage_name(self) -> str:
        return "domain"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: DomainInput,
    ) -> AsyncIterator[DomainResources]:
        """Compose domain dependencies in order.

        Args:
            stack: The dependency stack to use for composition.
            setup_input: Domain input containing infrastructure and component resources.

        Yields:
            DomainResources containing all domain dependencies.
        """
        # 1. Notification center (no dependencies)
        notification_center_dep = NotificationCenterDependency()
        notification_center = await stack.enter_dependency(notification_center_dep, None)

        # 2. Distributed lock factory (depends on config, db, etcd)
        distributed_lock_dep = DistributedLockFactoryDependency()
        distributed_lock_input = DistributedLockInput(
            config_provider=setup_input.config_provider,
            db=setup_input.db,
            etcd=setup_input.etcd,
        )
        distributed_lock_factory = await stack.enter_dependency(
            distributed_lock_dep, distributed_lock_input
        )

        # 3. Repositories (depends on db, storage_manager, config, valkey clients)
        repositories_dep = RepositoriesDependency()
        repositories_input = RepositoriesInput(
            db=setup_input.db,
            storage_manager=setup_input.storage_manager,
            config_provider=setup_input.config_provider,
            valkey_stat=setup_input.valkey_stat,
            valkey_live=setup_input.valkey_live,
            valkey_schedule=setup_input.valkey_schedule,
            valkey_image=setup_input.valkey_image,
        )
        repositories = await stack.enter_dependency(repositories_dep, repositories_input)

        # 4. Services context (depends on db)
        services_dep = ServicesContextDependency()
        services_input = ServicesInput(db=setup_input.db)
        services_ctx = await stack.enter_dependency(services_dep, services_input)

        yield DomainResources(
            notification_center=notification_center,
            distributed_lock_factory=distributed_lock_factory,
            repositories=repositories,
            services_ctx=services_ctx,
        )



"""Global middleware modules for the Manager REST API.

Re-exports commonly used middlewares so that ``server.py`` (and future
new-style modules) can import them from a single location.
"""

from ai.backend.common.metrics.http import build_api_metric_middleware
from ai.backend.common.middlewares.request_id import request_id_middleware

from .auth import build_auth_middleware
from .exception import build_exception_middleware

__all__ = (
    "build_api_metric_middleware",
    "build_auth_middleware",
    "build_exception_middleware",
    "request_id_middleware",
)



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.types import RepositoryArgs

from .repository import ManagerAdminRepository


@dataclass
class ManagerAdminRepositories:
    repository: ManagerAdminRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        return cls(repository=ManagerAdminRepository(db=args.db))



from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.prometheus_query_preset.actions.base import (
    PrometheusQueryPresetAction,
)


@dataclass
class DeletePresetAction(PrometheusQueryPresetAction):
    preset_id: UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.preset_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeletePresetActionResult(BaseActionResult):
    preset_id: UUID

    @override
    def entity_id(self) -> str | None:
        return None



"""migrate_image_data_to_rbac

Revision ID: 0e0723286a7a
Revises: 091d0274c2e3
Create Date: 2026-03-11 00:00:00.000000

"""

from uuid import UUID

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

from ai.backend.manager.models.rbac_models.migration.enums import (
    EntityType,
    OperationType,
)

# revision identifiers, used by Alembic.
revision = "0e0723286a7a"
down_revision = "091d0274c2e3"
branch_labels = None
depends_on = None

# Constants
BATCH_SIZE = 1000
MEMBER_ROLE_SUFFIX = "member"


def _add_entity_type_permissions(db_conn: Connection) -> None:
    """Add IMAGE entity-type permissions to all role+scope combinations.

    Uses a single set-based INSERT ... SELECT to derive IMAGE permissions
    for all role+scope combinations without application-side pagination.
    """
    # Precompute operation lists (sorted for deterministic ordering)
    member_ops = sorted(o.value for o in OperationType.member_operations())
    owner_ops = sorted(o.value for o in OperationType.owner_operations())

    insert_query = sa.text("""
        WITH role_scopes AS (
            SELECT DISTINCT
                p.role_id,
                r.name AS role_name,
                p.scope_type,
                p.scope_id
            FROM permissions p
            JOIN roles r ON p.role_id = r.id
        ),
        role_operations AS (
            -- Member operations for non-domain member roles
            SELECT
                rs.role_id,
                rs.scope_type,
                rs.scope_id,
                unnest(CAST(:member_ops AS text[])) AS operation
            FROM role_scopes rs
            WHERE rs.scope_type != 'domain'
              AND rs.role_name LIKE :member_pattern

            UNION ALL

            -- Owner operations for non-member roles
            SELECT
                rs.role_id,
                rs.scope_type,
                rs.scope_id,
                unnest(CAST(:owner_ops AS text[])) AS operation
            FROM role_scopes rs
            WHERE NOT (rs.role_name LIKE :member_pattern)
        )
        INSERT INTO permissions (role_id, scope_type, scope_id, entity_type, operation)
        SELECT
            role_id,
            scope_type,
            scope_id,
            :entity_type AS entity_type,
            operation
        FROM role_operations
        ON CONFLICT (role_id, scope_type, scope_id, entity_type, operation) DO NOTHING
    """)

    db_conn.execute(
        insert_query,
        {
            "member_ops": member_ops,
            "owner_ops": owner_ops,
            "member_pattern": f"%{MEMBER_ROLE_SUFFIX}",
            "entity_type": EntityType.IMAGE.value,
        },
    )


def _associate_image_to_scope(db_conn: Connection) -> None:
    """Associate image records to project scope via container registry.

    Pages by image IDs first, then JOINs to get all scope associations
    for each batch to avoid skipping rows in 1:N JOINs.
    """
    entity_type = EntityType.IMAGE.value
    relation_type = "auto"
    scope_type = "project"

    insert_query = sa.text("""
        INSERT INTO association_scopes_entities
            (scope_type, scope_id, entity_type, entity_id, relation_type)
        VALUES (:scope_type, :scope_id, :entity_type, :entity_id, :relation_type)
        ON CONFLICT (scope_type, scope_id, entity_id) DO NOTHING
    """)

    last_id = UUID("00000000-0000-0000-0000-000000000000")
    while True:
        id_query = sa.text("""
            SELECT id FROM images
            WHERE id > :last_id
            ORDER BY id
            LIMIT :limit
        """)
        id_rows = db_conn.execute(id_query, {"last_id": last_id, "limit": BATCH_SIZE}).all()
        if not id_rows:
            break

        last_id = id_rows[-1].id
        entity_ids = [row.id for row in id_rows]

        assoc_query = sa.text("""
            SELECT i.id AS entity_id, acrg.group_id AS scope_id
            FROM images i
            JOIN container_registries cr ON i.registry_id = cr.id
            JOIN association_container_registries_groups acrg ON cr.id = acrg.registry_id
            WHERE i.id = ANY(:entity_ids)
        """)
        rows = db_conn.execute(assoc_query, {"entity_ids": entity_ids}).all()

        values_list = [
            {
                "scope_type": scope_type,
                "scope_id": str(row.scope_id),
                "entity_type": entity_type,
                "entity_id": str(row.entity_id),
                "relation_type": relation_type,
            }
            for row in rows
        ]

        if values_list:
            db_conn.execute(insert_query, values_list)


def _remove_image_edges(db_conn: Connection) -> None:
    """Remove all IMAGE AUTO edges from association_scopes_entities."""
    entity_type = EntityType.IMAGE.value
    relation_type = "auto"

    while True:
        delete_query = sa.text("""
            DELETE FROM association_scopes_entities
            WHERE id IN (
                SELECT id FROM association_scopes_entities
                WHERE entity_type = :entity_type
                  AND relation_type = :relation_type
                LIMIT :limit
            )
        """)
        result = db_conn.execute(
            delete_query,
            {"entity_type": entity_type, "relation_type": relation_type, "limit": BATCH_SIZE},
        )
        if result.rowcount == 0:
            break


def _remove_image_permissions(db_conn: Connection) -> None:
    """Remove all IMAGE entity-type permissions."""
    entity_type = EntityType.IMAGE.value

    while True:
        delete_query = sa.text("""
            DELETE FROM permissions
            WHERE id IN (
                SELECT id FROM permissions
                WHERE entity_type = :entity_type
                LIMIT :limit
            )
        """)
        result = db_conn.execute(
            delete_query,
            {"entity_type": entity_type, "limit": BATCH_SIZE},
        )
        if result.rowcount == 0:
            break


def upgrade() -> None:
    conn = op.get_bind()
    _add_entity_type_permissions(conn)
    _associate_image_to_scope(conn)


def downgrade() -> None:
    conn = op.get_bind()
    _remove_image_edges(conn)
    _remove_image_permissions(conn)



"""GraphQL query resolvers for service catalog."""

from __future__ import annotations

from functools import lru_cache

import strawberry
from strawberry import Info

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only
from ai.backend.manager.models.service_catalog.row import ServiceCatalogRow
from ai.backend.manager.services.service_catalog.actions.search import SearchServiceCatalogsAction

from .types import ServiceCatalogFilterGQL, ServiceCatalogGQL, ServiceCatalogOrderByGQL

__all__ = ("admin_service_catalogs",)


@lru_cache(maxsize=1)
def _get_service_catalog_pagination_spec() -> PaginationSpec:
    """Get pagination spec for ServiceCatalog queries."""
    return PaginationSpec(
        forward_order=ServiceCatalogRow.service_group.asc(),
        backward_order=ServiceCatalogRow.service_group.desc(),
        forward_condition_factory=lambda cursor_value: lambda: ServiceCatalogRow.id > cursor_value,
        backward_condition_factory=lambda cursor_value: lambda: ServiceCatalogRow.id < cursor_value,
        tiebreaker_order=ServiceCatalogRow.id.asc(),
    )


@strawberry.field(  # type: ignore[misc]
    description="Added in 26.3.0. Query service catalog entries. Admin only.",
)
async def admin_service_catalogs(
    info: Info[StrawberryGQLContext],
    filter: ServiceCatalogFilterGQL | None = None,
    order_by: list[ServiceCatalogOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> list[ServiceCatalogGQL]:
    check_admin_only()
    ctx = info.context

    querier = ctx.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        _get_service_catalog_pagination_spec(),
        filter=filter,
        order_by=order_by,
    )

    action = SearchServiceCatalogsAction(querier=querier)
    result = await ctx.processors.service_catalog.search_service_catalogs.wait_for_complete(action)
    return [ServiceCatalogGQL.from_data(catalog) for catalog in result.data]
